#include <QMenu>
#include <QDir>
#include <QDesktopServices>
#include <QUrl>
#include <QKeyEvent>

#include "WidgetDownloadQueue.h"
#include "ui_WidgetDownloadQueue.h"
#include "DownloadQueueSelectionModel.h"

#include "nntpgrab_utils.h"

#ifdef MODEL_TEST
#include "modeltest/modeltest.h"
#endif

WidgetDownloadQueue::WidgetDownloadQueue(QNNTPGrabGlue *glue, QAction *actionRemoveSelectedFiles, QWidget *parent) :
        QWidget(parent),
        ui(new Ui::WidgetDownloadQueue)
{
    this->glue = glue;
    this->actionRemoveSelectedFiles = actionRemoveSelectedFiles;

    ui->setupUi(this);

    this->model = new DownloadQueueModel(glue);
#ifdef MODEL_TEST
    new ModelTest(this->model, this);
#endif
    ui->treeView->setModel(this->model);

    progressbarDelegate = new ProgressBarDelegateDownloadQueue(this);
    ui->treeView->setItemDelegateForColumn(DownloadQueueItem::FIELD_STATE_STR, progressbarDelegate);

    /* Connect the signal handlers for the download queue widgets */
    connect(ui->actionMoveSelectedItemsToTheTop, SIGNAL(triggered()), SLOT(btnQueueMoveToTop_clicked(void)));
    connect(ui->actionMoveSelectedItemsUp, SIGNAL(triggered()), SLOT(btnQueueMoveUp_clicked(void)));
    connect(ui->actionMoveSelectedItemsDown, SIGNAL(triggered()), SLOT(btnQueueMoveDown_clicked(void)));
    connect(ui->actionMoveSelectedItemsToTheBottom, SIGNAL(triggered()), SLOT(btnQueueMoveToBottom_clicked(void)));

    connect(ui->btnQueueMoveToTop, SIGNAL(clicked(void)), SLOT(btnQueueMoveToTop_clicked(void)));
    connect(ui->btnQueueMoveUp, SIGNAL(clicked(void)), SLOT(btnQueueMoveUp_clicked(void)));
    connect(ui->btnQueueMoveDown, SIGNAL(clicked(void)), SLOT(btnQueueMoveDown_clicked(void)));
    connect(ui->btnQueueMoveToBottom, SIGNAL(clicked(void)), SLOT(btnQueueMoveToBottom_clicked(void)));

    DownloadQueueSelectionModel *selectionModel = new DownloadQueueSelectionModel(this->model);
    ui->treeView->setSelectionModel(selectionModel);
    connect(selectionModel, SIGNAL(selectionChanged(const QItemSelection &, const QItemSelection &)), SLOT(onSelectionChanged(const QItemSelection &, const QItemSelection &)));

    connect(ui->actionMarkItemOptional, SIGNAL(triggered()), SLOT(actionMarkItemOptional_activated()));
    connect(ui->actionRestartSelectedItems, SIGNAL(triggered()), SLOT(actionRestartSelectedItems_activated()));
    connect(actionRemoveSelectedFiles, SIGNAL(triggered()), SLOT(actionRemoveSelectedFiles_activated()));
    connect(ui->actionOpenDownloadFolder, SIGNAL(triggered()), SLOT(actionOpenDownloadFolder_activated()));
    connect(ui->actionExecuteSelectedItem, SIGNAL(triggered()), SLOT(actionExecuteSelectedItem_activated()));

    ui->treeView->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(ui->treeView, SIGNAL(customContextMenuRequested(const QPoint &)), SLOT(showContextMenu(const QPoint &)));

#if QT_VERSION < 0x040600
    /* Moving files requires Qt 4.6.0 or higher */
    ui->btnQueueMoveToTop->hide();
    ui->btnQueueMoveUp->hide();
    ui->btnQueueMoveDown->hide();
    ui->btnQueueMoveToBottom->hide();
#endif

    /* Catch the DEL key */
    ui->treeView->installEventFilter(this);

    /* Add custom fields to the list */
    ui->treeView->setColumnHidden(DownloadQueueItem::FIELD_PROGRESS, true);
    ui->treeView->setColumnHidden(DownloadQueueItem::FIELD_FILE_SIZE, true);
    ui->treeView->setColumnHidden(DownloadQueueItem::FIELD_FILE_SIZE_REMAINING, true);
    ui->treeView->setColumnHidden(DownloadQueueItem::FIELD_REAL_FILENAME, true);

    /* Manually set the size of various columns */
    ui->treeView->setColumnWidth(DownloadQueueItem::FIELD_STATE_STR, 150);
    ui->treeView->setColumnWidth(DownloadQueueItem::FIELD_SUBJECT, 400);
    ui->treeView->setColumnWidth(DownloadQueueItem::FIELD_POSTER, 200);
    ui->treeView->setColumnWidth(DownloadQueueItem::FIELD_GROUPS, 150);
    ui->treeView->setColumnWidth(DownloadQueueItem::FIELD_FILE_SIZE_STR, 120);
    ui->treeView->setColumnWidth(DownloadQueueItem::FIELD_FILE_SIZE_REMAINING_STR, 120);
    ui->treeView->setColumnWidth(DownloadQueueItem::FIELD_STAMP, 210);
    ui->treeView->setColumnWidth(DownloadQueueItem::FIELD_ESTIMATED_TIME_REMAINING_STR, 230);
    ui->treeView->setColumnWidth(DownloadQueueItem::FIELD_ESTIMATED_TIME_TO_FINISH_STR, 200);

    /* Let the view indicate whenever a column resize is wanted */
    connect(model, SIGNAL(requestColumnResize(DownloadQueueItem::field)), SLOT(onRequestColumnResize(DownloadQueueItem::field)));
}

WidgetDownloadQueue::~WidgetDownloadQueue()
{
    delete progressbarDelegate;
    delete ui;
}

void WidgetDownloadQueue::changeEvent(QEvent *e)
{
    QWidget::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

bool WidgetDownloadQueue::eventFilter(QObject *watched, QEvent *e)
{
    if (watched == ui->treeView) {
        if (e->type() == QEvent::KeyPress) {
            QKeyEvent *keyEvent = static_cast<QKeyEvent*>(e);
            if (keyEvent->key() == Qt::Key_Delete) {
                actionRemoveSelectedFiles_activated();
                return true;
            }
        }
    }

    return QWidget::eventFilter(watched, e);
}

void WidgetDownloadQueue::doMoveItem(QString collection_name, QString subject, int new_position)
{
    if (subject.isNull()) {
        glue->schedularMoveCollection(collection_name, new_position);
        qDebug("Moved collection '%s' to position %i", collection_name.toUtf8().data(), new_position);
    } else {
        glue->schedularMoveTask(collection_name, subject, collection_name, new_position);
        qDebug("Moved subject '%s' of collection '%s' to position %i", subject.toUtf8().data(), collection_name.toUtf8().data(), new_position);
    }
}

int WidgetDownloadQueue::calculateNewPosition(int current_position, moveDirection direction)
{
    switch (direction) {
        case MOVE_DIRECTION_TOP:
            return 0;

        case MOVE_DIRECTION_UP:
            return qMax(0, current_position - 1);

        case MOVE_DIRECTION_DOWN:
            return current_position + 1;

        case MOVE_DIRECTION_BOTTOM:
            return -1;
    }

    return 0;
}

typedef struct _move_item_data {
    DownloadQueueItem *item;
    int position;
} MoveItemData;

void WidgetDownloadQueue::processSelectionAndMoveItems(moveDirection direction)
{
    DownloadQueueItem *collection_item;
    DownloadQueueItem *subject_item;
    QList<MoveItemData> list_collections;
    MoveItemData move_data;
    DownloadQueueModel *model = static_cast<DownloadQueueModel*>(ui->treeView->model());
    QItemSelectionModel *selectionModel = ui->treeView->selectionModel();

    for (int i = 0; i < model->rowCount(); i++) {
        QModelIndex idx_collection = model->index(i, 0);
        collection_item = static_cast<DownloadQueueItem*>(idx_collection.internalPointer());

        Q_ASSERT(collection_item != NULL);

        if (selectionModel->isSelected(idx_collection)) {
            if (i == 0 && (direction == MOVE_DIRECTION_UP || direction == MOVE_DIRECTION_TOP)) {
                /* We're already at the top, no need to move again */
                continue;
            }
            if (i == model->rowCount() - 1 && (direction == MOVE_DIRECTION_DOWN || direction == MOVE_DIRECTION_BOTTOM)) {
                /* We're already at the bottom, no need to move again */
                continue;
            }

            move_data.item = collection_item;
            move_data.position = i;

            if (direction == MOVE_DIRECTION_DOWN || direction == MOVE_DIRECTION_TOP) {
                list_collections.prepend(move_data);
            } else {
                list_collections.append(move_data);
            }

            continue;
        }

        QList<MoveItemData> list_subjects;

        for (int j = 0; j < collection_item->childCount(); j++) {
            subject_item = collection_item->child(j);
            QModelIndex idx_subject = model->index(subject_item->row(), 0, idx_collection);

            if (selectionModel->isSelected(idx_subject)) {
                if (j == 0 && (direction == MOVE_DIRECTION_UP || direction == MOVE_DIRECTION_TOP)) {
                    /* We're already at the top, no need to move again */
                    continue;
                }

                if (j == collection_item->childCount() - 1 && (direction == MOVE_DIRECTION_DOWN || direction == MOVE_DIRECTION_BOTTOM)) {
                    /* We're already at the bottom, no need to move again */
                    continue;
                }

                move_data.item = subject_item;
                move_data.position = j;

                if (direction == MOVE_DIRECTION_DOWN || direction == MOVE_DIRECTION_TOP) {
                    list_subjects.prepend(move_data);
                } else {
                    list_subjects.append(move_data);
                }

                continue;
            }
        }

        if (list_subjects.count() > 0) {
            Q_FOREACH(move_data, list_subjects) {
                doMoveItem(collection_item->subject, move_data.item->subject, calculateNewPosition(move_data.position, direction));
            }
        }
    }

    if (list_collections.count() > 0) {
        Q_FOREACH(move_data, list_collections) {
            doMoveItem(move_data.item->subject, NULL, calculateNewPosition(move_data.position, direction));
        }
    }
}

void WidgetDownloadQueue::btnQueueMoveToTop_clicked(void)
{
    processSelectionAndMoveItems(MOVE_DIRECTION_TOP);
}

void WidgetDownloadQueue::btnQueueMoveUp_clicked(void)
{
    processSelectionAndMoveItems(MOVE_DIRECTION_UP);
}

void WidgetDownloadQueue::btnQueueMoveDown_clicked(void)
{
    processSelectionAndMoveItems(MOVE_DIRECTION_DOWN);
}

void WidgetDownloadQueue::btnQueueMoveToBottom_clicked(void)
{
    processSelectionAndMoveItems(MOVE_DIRECTION_BOTTOM);
}

void WidgetDownloadQueue::onRequestColumnResize(DownloadQueueItem::field column)
{
    ui->treeView->resizeColumnToContents(column);
}

void WidgetDownloadQueue::onSelectionChanged(const QItemSelection &newSelection, const QItemSelection &oldSelection)
{
    Q_UNUSED(newSelection);
    Q_UNUSED(oldSelection);

    if (!ui->treeView->selectionModel()->hasSelection()) {
        ui->btnQueueMoveToTop->setEnabled(false);
        ui->btnQueueMoveUp->setEnabled(false);
        ui->btnQueueMoveDown->setEnabled(false);
        ui->btnQueueMoveToBottom->setEnabled(false);
        actionRemoveSelectedFiles->setEnabled(false);
    } else {
        ui->btnQueueMoveToTop->setEnabled(true);
        ui->btnQueueMoveUp->setEnabled(true);
        ui->btnQueueMoveDown->setEnabled(true);
        ui->btnQueueMoveToBottom->setEnabled(true);
        actionRemoveSelectedFiles->setEnabled(true);
    }
}

void WidgetDownloadQueue::actionMarkItemOptional_activated(void)
{
    QItemSelectionModel *selectionModel = ui->treeView->selectionModel();

    bool flag_set = FALSE;
    bool is_optional = FALSE;

    Q_FOREACH(QModelIndex idx, selectionModel->selectedRows()) {
        DownloadQueueItem *item = static_cast<DownloadQueueItem*>(idx.internalPointer());

        if (item->childCount() != 0) {
            /* This is a collection, ignore */
            continue;
        }

        /* This is a subject */
        if (!flag_set) {
            if (item->stateStr == tr("Skipped")) {
                is_optional = false;
            } else {
                is_optional = true;
            }

            flag_set = true;
        }

        if (!glue->schedularMarkTaskOptional(item->parent()->subject, item->subject, is_optional)) {
            qWarning("%s", tr("Unable to change the status of the selected file(s)").toUtf8().data());
        }
    }
}

void WidgetDownloadQueue::actionRemoveSelectedFiles_activated(void)
{
    QItemSelectionModel *selectionModel = ui->treeView->selectionModel();
    Q_FOREACH(QModelIndex idx, selectionModel->selectedRows()) {
        DownloadQueueItem *item = static_cast<DownloadQueueItem*>(idx.internalPointer());
        QString collection_name = "";
        QString subject = "";
        QString errmsg = "";

        if (item->childCount() == 0) {
            /* This is a subject */
            subject = item->subject;
            collection_name = item->parent()->subject;
        } else {
            /* This is a collection */
            collection_name = item->subject;
        }

        if (!glue->schedularDelTaskFromQueue(collection_name, subject, errmsg)) {
            qWarning(tr("Error occured while removing file(s):\n%s").toUtf8().data(), errmsg.toUtf8().data());
        }
    }
}

void WidgetDownloadQueue::actionRestartSelectedItems_activated(void)
{
    QItemSelectionModel *selectionModel = ui->treeView->selectionModel();
    Q_FOREACH(QModelIndex idx, selectionModel->selectedRows()) {
        DownloadQueueItem *item = static_cast<DownloadQueueItem*>(idx.internalPointer());
        QString collection_name = "";
        QString subject = "";
        QString errmsg = "";

        if (item->childCount() == 0) {
            /* This is a subject */
            subject = item->subject;
            collection_name = item->parent()->subject;
        } else {
            /* This is a collection */
            collection_name = item->subject;
        }

        if (!glue->schedularRestartTask(collection_name, subject, errmsg)) {
            qWarning(tr("Error occured while restarting file(s):\n%s").toUtf8().data(), errmsg.toUtf8().data());
        }
    }
}

void WidgetDownloadQueue::actionOpenDownloadFolder_activated(void)
{
    QString folder = "";
    QNNTPGrabGlue::ConfigOpts opts = glue->configGetOpts();

    if (ui->treeView->selectionModel()->selectedRows().count() == 1) {
        QString collection_name;
        DownloadQueueItem *item = static_cast<DownloadQueueItem*>(ui->treeView->selectionModel()->currentIndex().internalPointer());

        if (item->childCount() == 0) {
            /* This is a subject */
            collection_name = item->parent()->subject;
        } else {
            collection_name = item->subject;
        }

        folder = opts.download_directory + QString("/") + collection_name;
    } else {
        folder = opts.download_directory;
    }

    QFileInfo location(folder);
    if (!location.exists()) {
        /* Try to parent folder */
        folder = opts.download_directory;
        location.setFile(folder);
        if (!location.exists()) {
            qWarning(tr("Folder '%s' does not exist").toUtf8().data(), location.absoluteFilePath().toUtf8().data());
            return;
        }
    }

#ifdef Q_OS_WIN32
    QDesktopServices::openUrl(QUrl(QString("file:///") + location.absoluteFilePath()));
#else
    QDesktopServices::openUrl(QUrl(QString("file://") + location.absoluteFilePath()));
#endif
}

void WidgetDownloadQueue::actionExecuteSelectedItem_activated(void)
{
    if (ui->treeView->selectionModel()->selectedRows().count() != 1) {
        return;
    }

    DownloadQueueItem *item = static_cast<DownloadQueueItem*>(ui->treeView->selectionModel()->currentIndex().internalPointer());

    if (item->childCount() != 0) {
        return;
    }

    QString collection_name = item->parent()->subject;
    QString real_filename = item->realFilename;

    if (real_filename.isEmpty()) {
        qWarning("%s", tr("Unable to execute file as it isn't downloaded yet").toUtf8().data());
        return;
    }

    QNNTPGrabGlue::ConfigOpts opts = glue->configGetOpts();
    QFileInfo location(QString(opts.download_directory) + QString("/") + collection_name + QString("/") + real_filename);

    if (!location.exists()) {
        qWarning(tr("File '%s' does not exist").toUtf8().data(), location.absoluteFilePath().toUtf8().data());
        return;
    }

#ifdef Q_OS_WIN32
    QDesktopServices::openUrl(QUrl(QString("file:///") + location.absoluteFilePath()));
#else
    QDesktopServices::openUrl(QUrl(QString("file://") + location.absoluteFilePath()));
#endif
}

void WidgetDownloadQueue::showContextMenu(const QPoint &position)
{
    QList<QAction *> actions;
    QAction *a = new QAction(this);

    if (!ui->treeView->indexAt(position).isValid()) {
        return;
    }

    if (!ui->treeView->selectionModel()->hasSelection()) {
        return;
    }

    actions.append(this->actionRemoveSelectedFiles);
    actions.append(ui->actionMarkItemOptional);
    actions.append(ui->actionRestartSelectedItems);
    actions.append(ui->actionOpenDownloadFolder);
    actions.append(ui->actionExecuteSelectedItem);

    a->setSeparator(true);
    actions.append(a);
    actions.append(ui->actionMoveSelectedItemsToTheTop);
    actions.append(ui->actionMoveSelectedItemsUp);
    actions.append(ui->actionMoveSelectedItemsDown);
    actions.append(ui->actionMoveSelectedItemsToTheBottom);

    if (!glue->getIsStandalone()) {
        ui->actionOpenDownloadFolder->setEnabled(false);
    } else {
        ui->actionOpenDownloadFolder->setEnabled(true);
    }

    /* Do we have exactly one selected item and is this item already downloaded? */
    ui->actionExecuteSelectedItem->setEnabled(false);
    DownloadQueueItem *item = static_cast<DownloadQueueItem*>(ui->treeView->selectionModel()->currentIndex().internalPointer());
    if (glue->getIsStandalone() && ui->treeView->selectionModel()->selectedRows().count() == 1) {
        if (item->stateStr == tr("Done") ||
            item->stateStr == tr("Incomplete")) {

            ui->actionExecuteSelectedItem->setEnabled(true);
        }
    }

    /* Enable the forcefully download / skip option */
    if (item->stateStr == tr("Skipped")) {
        ui->actionMarkItemOptional->setText(tr("Forcefully download file(s)"));
        ui->actionMarkItemOptional->setEnabled(true);
    } else if (item->stateStr != tr("Done") &&
               item->stateStr != tr("Incomplete") &&
               item->stateStr != tr("Not available")) {

        ui->actionMarkItemOptional->setText(tr("Don't download file(s)"));
        ui->actionMarkItemOptional->setEnabled(true);
    } else {
        ui->actionMarkItemOptional->setEnabled(false);
    }

    QMenu::exec(actions, ui->treeView->mapToGlobal(position));

    delete a;
}

void WidgetDownloadQueue::hideField(DownloadQueueItem::field f)
{
    ui->treeView->hideColumn(f);
}

void WidgetDownloadQueue::showField(DownloadQueueItem::field f)
{
    ui->treeView->showColumn(f);
}
