/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <glib/gstdio.h>

#include "nntpgrab.h"
#include "collections.h"
#include "collection_alloc.h"
#include "queue_rawfile.h"

static GStaticMutex db_mutex = G_STATIC_MUTEX_INIT;
static char *current_mode = NULL;

static char *
generate_filename(void)
{
    if (g_getenv("NNTPGRAB_CONFIG_DIR")) {
        return g_build_filename(g_getenv("NNTPGRAB_CONFIG_DIR"), "NNTPGrab", "download_queue.db", NULL);
    } else {
        return g_build_filename(g_get_user_config_dir(), "NNTPGrab", "download_queue.db", NULL);
    }
}

static FILE *
queue_rawfile_open(const char *mode, char **errmsg)
{
    FILE *fp;
    char *filename;
    char *dirname;

    /* Prevent another thread from claiming the database */
    g_static_mutex_lock(&db_mutex);

    /* Generate the folder ~/.config/NNTPGrab in case it didn't exist yet */
    filename = generate_filename();
    dirname = g_path_get_dirname(filename);
    g_mkdir_with_parents(dirname, 0700);
    g_free(dirname);

    /* When opening in write mode, don't overwrite the original file yet, but create a new (temporary) file */
    if (!strcmp(mode, "w")) {
        char *tmp = g_strdup_printf("%s%s", filename, ".new");
        g_free(filename);
        filename = tmp;
        current_mode = g_strdup(mode);
    } else {
        current_mode = NULL;
    }

    fp = g_fopen(filename, mode);
    if (!fp) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Unable to open download queue '%s': %s"), filename, strerror(errno));
        } else if (mode[0] != 'r') {
            g_warning(_("Unable to open download queue '%s': %s"), filename, strerror(errno));
        }

        g_free(filename);

        g_static_mutex_unlock(&db_mutex);

        return NULL;
    }

    g_free(filename);

    return fp;
}

static gboolean
queue_rawfile_close(FILE *fp, char **errmsg)
{
    gboolean ret = TRUE;

    g_return_val_if_fail(fp != NULL, FALSE);

    if (fclose(fp) != 0) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("Error occured while closing download queue file: %s"), strerror(errno));
        } else {
            g_warning(_("Error occured while closing download queue file: %s"), strerror(errno));
        }

        ret = FALSE;
    }

    /* If the file was opened in write mode, rename the tempory file to the correct filename */
    if (current_mode) {
        if (!strcmp(current_mode, "w") && ret == TRUE) {
            char *filename = generate_filename();
            char *tmp = g_strdup_printf("%s%s", filename, ".new");

            if (g_rename(tmp, filename) == -1) {
#ifdef WIN32
                /* The rename can fail on Win32 if the dest file already exists or some virus scanner is claiming the file */
                /* Try removing the dest file first and try to rename again afterwards */
                if (!g_file_test(filename, G_FILE_TEST_EXISTS)) {
                    if (errmsg) {
                        *errmsg = g_strdup_printf(_("Unable to rename temporary download queue file '%s' to '%s': %s"), tmp, filename, strerror(errno));
                    } else {
                        g_warning(_("Unable to rename temporary download queue file '%s' to '%s': %s"), tmp, filename, strerror(errno));
                    }

                    ret = FALSE;
                } else {
                    /* Dest file exists, try to remove the dest file */
                    if (g_unlink(filename) == -1) {
                        if (errmsg) {
                            *errmsg = g_strdup_printf(_("Unable to remove existing download queue '%s': %s"), filename, strerror(errno));
                        } else {
                            g_warning(_("Unable to remove existing download queue '%s': %s"), filename, strerror(errno));
                        }

                        ret = FALSE;
                    } else {
                        /* Remove failed, try to rename again */
                        if (g_rename(tmp, filename) == -1) {
                            if (errmsg) {
                                *errmsg = g_strdup_printf(_("Unable to rename temporary download queue file '%s' to '%s': %s"), tmp, filename, strerror(errno));
                            } else {
                                g_warning(_("Unable to rename temporary download queue file '%s' to '%s': %s"), tmp, filename, strerror(errno));
                            }

                            ret = FALSE;
                        }
                    }
                }
#else
                if (errmsg) {
                    *errmsg = g_strdup_printf(_("Unable to rename temporary download queue file '%s' to '%s': %s"), tmp, filename, strerror(errno));
                } else {
                    g_warning(_("Unable to rename temporary download queue file '%s' to '%s': %s"), tmp, filename, strerror(errno));
                }

                ret = FALSE;
#endif
            }
            g_free(tmp);
            g_free(filename);
        }

        g_free(current_mode);
        current_mode = NULL;
    }

    g_static_mutex_unlock(&db_mutex);

    return ret;
}

gboolean
queue_rawfile_load(GList **queue_ret, char **errmsg)
{
    FILE *fp;
    char line[65535];
    gboolean retval = TRUE;
    NNTPCollection *active_collection = NULL;
    NNTPFile *active_file = NULL;
    char **values;

    g_return_val_if_fail(queue_ret != NULL, FALSE);

    *queue_ret = NULL;

    /* Open a file pointer where all data can be read */
    if (!(fp = queue_rawfile_open("r", errmsg))) {
        return FALSE;
    }

    while (!feof(fp)) {
        memset(&line, 0, sizeof(line));

        if (!fgets(line, sizeof(line) - 1, fp)) {
            if (feof(fp)) {
                break;
            }

            if (errmsg) {
                *errmsg = g_strdup_printf(_("Unknown error occured while reading the download queue: %s"), strerror(errno));
            } else {
                g_warning(_("Unknown error occured while reading the download queue: %s"), strerror(errno));
            }

            retval = FALSE;
            break;
        }

        /* Strip out any newlines */
        if (line[strlen(line) - 1] == '\n') {
            line[strlen(line) - 1] = '\0';
        }
        if (line[strlen(line) - 1] == '\r') {
            line[strlen(line) - 1] = '\0';
        }

        /* Find out the type of the line
         * No tabs is a collection
         * One tabs is a file
         * Two tabs is a part
         */
        if (line[0] == '\t' && line[1] == '\t') {
            /* Part detected */
            NNTPPart *part;
            GList *parts;

            if (active_collection == NULL || active_file == NULL) {
                goto malformed_file;
            }

            values = g_strsplit(line, "\t", -1);
            if (!values || !values[0] || !values[1] || !values[2] || !values[3] || !values[4] || !values[5]) {
                goto malformed_file;
            }

            part = g_slice_new0(NNTPPart);
            strncpy((char*) part->message_id, values[2], sizeof(part->message_id) -1);
            part->size = atoi(values[3]);
            part->part_num = atoi(values[4]);
            part->downloaded = atoi(values[5]);

            g_strfreev(values);

            active_file->parts = g_list_append(active_file->parts, part);
            active_file->num_parts++;
            active_file->file_size += part->size;
            active_collection->total_size += part->size;
            if (part->downloaded) {
                active_file->num_parts_downloaded++;
            } else {
                active_file->file_size_remaining += part->size;
                if (active_file->status != TASK_STATE_SKIPPED) {
                    active_collection->total_size_remaining += part->size;
                } else {
                    active_collection->total_size -= part->size;
                }
            }

            parts = active_file->parts;
            active_file->file_is_downloaded = TRUE;
            while (parts) {
                NNTPPart *part = parts->data;

                if (!part->downloaded) {
                    active_file->file_is_downloaded = FALSE;
                }

                parts = g_list_next(parts);
            }

            switch (active_file->status) {
                case TASK_STATE_DOWNLOADING:
                    active_file->file_is_downloading = TRUE;
                    break;

                case TASK_STATE_WAITING_FOR_DECODE:
                case TASK_STATE_DECODING:
                    active_file->file_is_downloaded = FALSE;
                    active_file->file_is_downloading = TRUE;    /* Needed to requeue the decoding */
                    break;

                case TASK_STATE_WAITING_FOR_DOWNLOAD:
                    active_file->file_is_downloaded = FALSE;
                    break;

                case TASK_STATE_FINISHED_COMPLETE:
                case TASK_STATE_FINISHED_INCOMPLETE:
                case TASK_STATE_FINISHED_NO_PARTS_AVAIL:
                case TASK_STATE_SKIPPED:
                    break;
            }
        } else if (line[0] == '\t') {
            /* File detected */
            gboolean file_lacks_md5_field = FALSE;
            char **groups_values;
            int i;

            if (active_collection == NULL) {
                goto malformed_file;
            }

            values = g_strsplit(line, "\t", -1);
            if (!values || !values[0] || !values[1] || !values[2] || !values[3] || !values[4] || !values[5] ||
                           !values[6] || !values[7] || !values[8] || !values[9] || !values[10] || !values[11]) {

                goto malformed_file;
            }

            if (!values[12]) {
                file_lacks_md5_field = TRUE;
            }

            active_file = file_new();
            strncpy((char*) active_file->subject, values[1], sizeof(active_file->subject) - 1);
            strncpy((char*) active_file->poster, values[2], sizeof(active_file->poster) - 1);
            active_file->stamp = atoi(values[3]);
            strncpy((char*) active_file->tmp_filename, values[4], sizeof(active_file->tmp_filename) - 1);
            strncpy((char*) active_file->real_filename, values[5], sizeof(active_file->real_filename) - 1);
            active_file->status = atoi(values[6]);
            active_file->file_type = atoi(values[7]);
            active_file->par2_start_num = atoi(values[8]);
            active_file->par2_end_num = atoi(values[9]);
            active_file->par2_num_blocks = atoi(values[10]);

            if (file_lacks_md5_field) {
                groups_values = g_strsplit(values[11], ",", -1);
            } else {
                strncpy((char*) active_file->md5sum, values[11], sizeof(active_file->md5sum) - 1);
                groups_values = g_strsplit(values[12], ",", -1);
            }

            i = 0;
            while (groups_values[i]) {
                active_file->groups = g_list_append(active_file->groups, g_strdup(groups_values[i]));
                i++;
            }

            g_strfreev(groups_values);
            g_strfreev(values);

            active_collection->files = g_list_append(active_collection->files, active_file);
        } else if (line[0] != '\t') {
            /* Collection detected */

            values = g_strsplit(line, "\t", -1);
            if (!values || !values[0] || !values[1] || !values[2]) {
                goto malformed_file;
            }

            active_collection = collection_new();
            strncpy((char*) active_collection->collection_name, values[0], sizeof(active_collection->collection_name) - 1);
            strncpy((char*) active_collection->poster, values[1], sizeof(active_collection->poster) - 1);
            strncpy((char*) active_collection->groups, values[2], sizeof(active_collection->groups) - 1);

            g_strfreev(values);

            *queue_ret = g_list_append(*queue_ret, active_collection);
        }
    }

out:
    /* Close the file. Error checking isn't necessary here as we've only read something */
    queue_rawfile_close(fp, NULL);

    return retval;

malformed_file:
    if (errmsg) {
        *errmsg = g_strdup_printf(_("Malformed download queue detected"));
    } else {
        g_warning(_("Malformed download queue detected"));
    }

    goto out;
}

gboolean
queue_rawfile_save(GList *queue, char **errmsg)
{
    GList *list = queue;
    FILE *fp;
    char groups[1024];

    /* Open a file pointer where all data can be saved */
    if (!(fp = queue_rawfile_open("w", errmsg))) {
        return FALSE;
    }

    /* Collect all data */
    while (list) {
        NNTPCollection *collection = list->data;
        GList *list2;

        fprintf(fp, "%s\t%s\t%s\n", collection->collection_name, collection->poster, collection->groups);

        list2 = collection->files;
        while (list2) {
            NNTPFile *file = list2->data;
            GList *list3;

            memset(groups, 0, sizeof(groups));
            list3 = file->groups;
            while (list3) {
                const char *groupname = list3->data;

                if (strlen(groups) + strlen(groupname) > sizeof(groups) - 2) {
                    /* Prevent a buffer overflow, limit the number of groups in posting which are crossposted to a large number of groups */
                    list3 = g_list_next(list3);
                    continue;
                }

                if (strlen(groups) > 0) {
                    strcat(groups, ",");
                }

                strcat(groups, groupname);

                list3 = g_list_next(list3);
            }

            fprintf(fp, "\t%s\t%s\t%i\t%s\t%s\t%i\t%i\t%i\t%i\t%i\t%s\t%s\n", file->subject, file->poster, (int) file->stamp, file->tmp_filename, file->real_filename, file->status, file->file_type, file->par2_start_num, file->par2_end_num, file->par2_num_blocks, file->md5sum, groups);

            list3 = file->parts;
            while (list3) {
                NNTPPart *part = list3->data;

                fprintf(fp, "\t\t%s\t%i\t%i\t%i\n", part->message_id, part->size, part->part_num, part->downloaded);

                list3 = g_list_next(list3);
            }

            list2 = g_list_next(list2);
        }

        list = g_list_next(list);
    }

    /* Close the file and make sure that everything got written back to disk */
    if (!queue_rawfile_close(fp, errmsg)) {
        return FALSE;
    }

    return TRUE;
}
